#!/usr/bin/env python3
# Author:  Octavio Castillo Reyes
# Contact: octavio.castillo@bsc.es
"""Define data preprocessing operations for **PETGEM**."""


# ---------------------------------------------------------------
# Load python modules
# ---------------------------------------------------------------
import numpy as np
import h5py
from scipy.spatial import Delaunay
from petsc4py import PETSc

# ---------------------------------------------------------------
# Load petgem modules (BSC)
# ---------------------------------------------------------------
from .common import Print, Timers, measure_all_class_methods
from .parallel import MPIEnvironment, createSequentialDenseMatrixWithArray
from .parallel import writeParallelDenseMatrix, createSequentialVectorWithArray
from .parallel import writePetscVector
from .mesh import readGmshNodes, readGmshConnectivity, computeEdges, computeFaces
from .mesh import computeBoundaryFaces, computeBoundaryEdges, computeBoundaries
from .mesh import readGmshPhysicalGroups
from .vectors import invConnectivity
from .hvfem import computeConnectivityDOFS

# ###############################################################
# ################     CLASSES DEFINITION      ##################
# ###############################################################
@measure_all_class_methods
class Preprocessing():
    """Class for preprocessing."""

    def __init__(self):
        """Initialization of a preprocessing class."""
        return

    def run(self, setup):
        """Run a preprocessing task.

        :param obj setup: inputSetup object.
        :return: None
        """
        # ---------------------------------------------------------------
        # Initialization
        # ---------------------------------------------------------------
        # Start timer
        Timers()["Preprocessing"].start()

        # Parameters shortcut (for code legibility)
        model = setup.model
        output = setup.output

        # Obtain the MPI environment
        parEnv = MPIEnvironment()

        # ---------------------------------------------------------------
        # Import mesh file (gmsh format)
        # ---------------------------------------------------------------
        # Read nodes
        nodes, _ = readGmshNodes(model.mesh_file)

        # Read connectivity
        elemsN, nElems = readGmshConnectivity(model.mesh_file)

        # ---------------------------------------------------------------
        # Preprocessing nodal coordinates
        # ---------------------------------------------------------------
        Print.master('     Nodal coordinates')

        # Build coordinates in PETGEM format where each row
        # represent the xyz coordinates of the 4 tetrahedral element
        num_dimensions = 3
        num_nodes_per_element = 4
        data = np.array((nodes[elemsN[:], :]), dtype=np.float)
        data = data.reshape(nElems, num_dimensions*num_nodes_per_element)

        # Get matrix dimensions
        size = data.shape

        # Build PETSc structures
        matrix = createSequentialDenseMatrixWithArray(size[0], size[1], data)

        # Build path to save the file
        out_path = output.directory_scratch + '/nodes.dat'

        if parEnv.rank == 0:
            # Write PETGEM nodes in PETSc format
            writeParallelDenseMatrix(out_path, matrix, communicator=PETSc.COMM_SELF)

        # ---------------------------------------------------------------
        # Preprocessing mesh connectivity
        # ---------------------------------------------------------------
        Print.master('     Mesh connectivity')

        # Get matrix dimensions
        size = elemsN.shape

        # Build PETSc structures
        matrix = createSequentialDenseMatrixWithArray(size[0], size[1], elemsN)

        # Build path to save the file
        out_path = output.directory_scratch + '/meshConnectivity.dat'

        if parEnv.rank == 0:
            # Write PETGEM connectivity in PETSc format
            writeParallelDenseMatrix(out_path, matrix, communicator=PETSc.COMM_SELF)

        # ---------------------------------------------------------------
        # Preprocessing edges connectivity
        # ---------------------------------------------------------------
        Print.master('     Edges connectivity')

        # Compute edges
        elemsE, edgesNodes = computeEdges(elemsN, nElems)
        nEdges = edgesNodes.shape[0]

        # Get matrix dimensions
        size = elemsE.shape

        # Build PETSc structures
        matrix = createSequentialDenseMatrixWithArray(size[0], size[1], elemsE)

        # Build path to save the file
        out_path = output.directory_scratch + '/edges.dat'

        if parEnv.rank == 0:
            # Write PETGEM edges in PETSc format
            writeParallelDenseMatrix(out_path, matrix, communicator=PETSc.COMM_SELF)

        # Reshape edgesNodes and save
        num_nodes_per_edge = 2
        num_edges_per_element = 6
        data = np.array((edgesNodes[elemsE[:], :]), dtype=np.float)
        data = data.reshape(nElems, num_nodes_per_edge*num_edges_per_element)

        # Get matrix dimensions
        size = data.shape

        # Build PETSc structures
        matrix = createSequentialDenseMatrixWithArray(size[0], size[1], data)

        # Build path to save the file
        out_path = output.directory_scratch + '/edgesNodes.dat'

        if parEnv.rank == 0:
            # Write PETGEM edgesNodes in PETSc format
            writeParallelDenseMatrix(out_path, matrix, communicator=PETSc.COMM_SELF)

        # ---------------------------------------------------------------
        # Preprocessing faces connectivity
        # ---------------------------------------------------------------
        Print.master('     Faces connectivity')

        # Compute faces
        elemsF, facesN = computeFaces(elemsN, nElems)
        nFaces = facesN.shape[0]

        # Get matrix dimensions
        size = elemsF.shape
        # Build PETSc structures
        matrix = createSequentialDenseMatrixWithArray(size[0], size[1], elemsF)

        # Build path to save the file
        out_path = output.directory_scratch + '/faces.dat'

        if parEnv.rank == 0:
            # Write PETGEM edges in PETSc format
            writeParallelDenseMatrix(out_path, matrix, communicator=PETSc.COMM_SELF)

        # ---------------------------------------------------------------
        # Preprocessing faces-edges connectivity
        # ---------------------------------------------------------------
        Print.master('     Faces-edges connectivity')

        N = invConnectivity(elemsF, nFaces)

        if nElems != 1:
            N = np.delete(N, 0, axis=1)

        # Allocate
        facesE = np.zeros((nFaces, 3), dtype=np.int)

        # Compute edges list for each face
        for i in np.arange(nFaces):
            iEle = N[i, 0]
            edgesEle = elemsE[iEle,:]
            facesEle = elemsF[iEle,:]
            kFace = np.where(facesEle == i)[0]
            if kFace == 0:  # Face 1
                facesE[facesEle[kFace],:] = [edgesEle[0], edgesEle[1], edgesEle[2]]
            elif kFace == 1:  # Face 2
                facesE[facesEle[kFace],:] = [edgesEle[0], edgesEle[4], edgesEle[3]]
            elif kFace == 2:  # Face 3
                facesE[facesEle[kFace],:] = [edgesEle[1], edgesEle[5], edgesEle[4]]
            elif kFace == 3:  # Face 4
                facesE[facesEle[kFace],:] = [edgesEle[2], edgesEle[5], edgesEle[3]]

        num_faces_per_element = 4
        num_edges_per_face = 3
        data = np.array((facesE[elemsF[:], :]), dtype=np.float)
        data = data.reshape(nElems, num_faces_per_element*num_edges_per_face)

        # Get matrix dimensions
        size = data.shape

        # Build PETSc structures
        matrix = createSequentialDenseMatrixWithArray(size[0], size[1], data)

        # Build path to save the file
        out_path = output.directory_scratch + '/facesEdges.dat'

        if parEnv.rank == 0:
            # Write PETGEM edges in PETSc format
            writeParallelDenseMatrix(out_path, matrix, communicator=PETSc.COMM_SELF)

        # ---------------------------------------------------------------
        # Preprocessing dofs connectivity
        # ---------------------------------------------------------------
        Print.master('     DOFs connectivity')

        # Compute degrees of freedom connectivity
        dofs, dof_edges, dof_faces, _, total_num_dofs = computeConnectivityDOFS(elemsE,elemsF,model.basis_order)

        # Get matrix dimensions
        size = dofs.shape

        # Build PETSc structures
        matrix = createSequentialDenseMatrixWithArray(size[0], size[1], dofs)

        # Build path to save the file
        out_path = output.directory_scratch + '/dofs.dat'

        if parEnv.rank == 0:
            # Write PETGEM edges in PETSc format
            writeParallelDenseMatrix(out_path, matrix, communicator=PETSc.COMM_SELF)

        # ---------------------------------------------------------------
        # Preprocessing boundaries
        # ---------------------------------------------------------------
        Print.master('     Boundaries')

        # Compute boundary faces
        bFacesN, bFaces = computeBoundaryFaces(elemsF, facesN)

        # Compute boundary edges
        bEdges = computeBoundaryEdges(edgesNodes, bFacesN)

        # Compute dofs on boundaries
        _, indx_boundary_dofs = computeBoundaries(dofs, dof_edges, dof_faces, bEdges, bFaces, model.basis_order);

        # Build PETSc structures
        vector = createSequentialVectorWithArray(indx_boundary_dofs)

        # Build path to save the file
        out_path = output.directory_scratch + '/boundaries.dat'

        if parEnv.rank == 0:
            # Write PETGEM nodes in PETSc format
            writePetscVector(out_path, vector, communicator=PETSc.COMM_SELF)

        # ---------------------------------------------------------------
        # Preprocessing sigma model
        # ---------------------------------------------------------------
        Print.master('     Conductivity model')

        # Read element's tag
        elemsS, nElems = readGmshPhysicalGroups(model.mesh_file)

        # Build conductivity arrays
        conductivityModel = np.zeros((nElems, 2), dtype=np.float)
        for i in np.arange(nElems):
            # Set horizontal sigma
            conductivityModel[i, 0] = model.sigma_horizontal[np.int(elemsS[i])]
            # Set vertical sigma
            conductivityModel[i, 1] = model.sigma_vertical[np.int(elemsS[i])]

        # Get matrix dimensions
        size = conductivityModel.shape

        # Build PETSc structures
        matrix = createSequentialDenseMatrixWithArray(size[0], size[1], conductivityModel)

        # Build path to save the file
        out_path = output.directory_scratch + '/conductivityModel.dat'

        if parEnv.rank == 0:
            # Write PETGEM edges in PETSc format
            writeParallelDenseMatrix(out_path, matrix, communicator=PETSc.COMM_SELF)

        # ---------------------------------------------------------------
        # Preprocessing receivers
        # ---------------------------------------------------------------
        Print.master('     Receivers')

        # Open receivers_file
        fileID = h5py.File(model.receivers_file, 'r')

        # Read receivers
        receivers = fileID.get('data')[()]

        # Number of receivers
        if receivers.ndim == 1:
            nReceivers = 1
        else:
            dim = receivers.shape
            nReceivers = dim[0]

        # Build Delaunay triangulation with nodes
        tri = Delaunay(nodes)

        # Overwrite Delaunay structure with mesh_file connectivity and points
        tri.simplices = elemsN.astype(np.int32)
        tri.vertices = elemsN.astype(np.int32)

        # Find out which tetrahedral element points are in
        recvElems = tri.find_simplex(receivers, bruteforce=True, tol=1.e-12)

        # Find out which tetrahedral element source point is in
        srcElem = tri.find_simplex(model.src_position, bruteforce=True, tol=1.e-12)

        # Determine if all receiver points were found
        idx = np.where(np.logical_or(recvElems>nElems, recvElems<0))[0]

        # If idx is not empty, there are receivers outside the domain
        if idx.size != 0:
            Print.master('        The following receivers were not located and will not be taken into account ' + str(idx))
            # Update number of receivers
            nReceivers = nReceivers - len(idx)

            if nReceivers == 0:
                Print.master('     No receiver has been found. Nothing to do. Aborting')
                exit(-1)

            # Remove idx from receivers matrix
            receivers = np.delete(receivers, idx, axis=0)

            # Remove idx from recvElems
            recvElems = np.delete(recvElems, idx, axis=0)

        # If srcElem is empty, source not located
        if srcElem == 0:
            Print.master('        Source no located in the computational domain. Please, improve the mesh quality')
            exit(-1)

        # Compute number of dofs per element
        num_dof_in_element = np.int(model.basis_order*(model.basis_order+2)*(model.basis_order+3)/2)

        # Allocate
        data_receiver = np.zeros((nReceivers, 53+num_dof_in_element), dtype=np.float)

        # Fill tmp matrix with receiver positions, element coordinates and
        # nodal indexes
        for i in np.arange(nReceivers):
            # If there is one receiver
            if nReceivers == 1:
                # Get index of tetrahedral element (receiver container)
                iEle = recvElems
                # Get dofs of element container
                dofsElement = dofs[iEle]
            # If there are more than one receivers
            else:
                # Get index of tetrahedral element (receiver container)
                iEle = recvElems[i]
                # Get dofs of element container
                dofsElement = dofs[iEle, :]

            # Get indexes of nodes for iand insert
            nodesReceiver = elemsN[iEle, :]
            data_receiver[i, 0:4] = nodesReceiver
            # Get nodes coordinates for i and insert
            coordEle = nodes[nodesReceiver, :]
            coordEle = coordEle.flatten()
            data_receiver[i, 4:16] = coordEle
            # Get indexes of faces for i and insert
            facesReceiver = elemsF[iEle, :]
            data_receiver[i, 16:20] = facesReceiver
            # Get edges indexes for faces in i and insert
            edgesReceiver = facesE[facesReceiver, :]
            edgesReceiver = edgesReceiver.flatten()
            data_receiver[i, 20:32] = edgesReceiver
            # Get indexes of edges for i and insert
            edgesReceiver = elemsE[iEle, :]
            data_receiver[i, 32:38] = edgesReceiver
            # Get node indexes for edges in i and insert
            edgesNodesReceiver = edgesNodes[edgesReceiver, :]
            edgesNodesReceiver = edgesNodesReceiver.flatten()
            data_receiver[i, 38:50] = edgesNodesReceiver
            # Get receiver coordinates
            coordReceiver = receivers[i,: ]
            data_receiver[i, 50:53] = coordReceiver
            # Get dofs for srcElem and insert
            dofsReceiver = dofsElement
            data_receiver[i, 53::] = dofsReceiver

        # Get matrix dimensions
        size = data_receiver.shape

        # Build PETSc structures
        matrix = createSequentialDenseMatrixWithArray(size[0], size[1], data_receiver)

        # Build path to save the file
        out_path = output.directory_scratch + '/receivers.dat'

        if parEnv.rank == 0:
            # Write PETGEM receivers in PETSc format
            writeParallelDenseMatrix(out_path, matrix, communicator=PETSc.COMM_SELF)

        # Compute number of dofs per element
        num_dof_in_element = np.int(model.basis_order*(model.basis_order+2)*(model.basis_order+3)/2)

        # Build data for source insertion
        vector = np.zeros(50+num_dof_in_element, dtype=np.float)

        # Get indexes of nodes for srcElem and insert
        nodesSource = elemsN[srcElem, :]
        vector[0:4] = nodesSource
        # Get nodes coordinates for srcElem and insert
        coordSource = nodes[nodesSource, :]
        coordSource = coordSource.flatten()
        vector[4:16] = coordSource
        # Get indexes of faces for srcElem and insert
        facesSource = elemsF[srcElem, :]
        vector[16:20] = facesSource
        # Get edges indexes for faces in srcElem and insert
        edgesFace = facesE[facesSource, :]
        edgesFace = edgesFace.flatten()
        vector[20:32] = edgesFace
        # Get indexes of edges for srcElem and insert
        edgesSource = elemsE[srcElem, :]
        vector[32:38] = edgesSource
        # Get node indexes for edges in srcElem and insert
        edgesNodesSource = edgesNodes[edgesSource, :]
        edgesNodesSource = edgesNodesSource.flatten()
        vector[38:50] = edgesNodesSource
        # Get dofs for srcElem and insert
        dofsSource = dofs[srcElem,:]
        vector[50::] = dofsSource

        # Build PETSc structures
        vector = createSequentialVectorWithArray(vector)

        # Build path to save the file
        out_path = output.directory_scratch + '/source.dat'

        if parEnv.rank == 0:
            # Write PETGEM nodes in PETSc format
            writePetscVector(out_path, vector, communicator=PETSc.COMM_SELF)

        # ---------------------------------------------------------------
        # Sparsity pattern
        # ---------------------------------------------------------------
        # Setup valence for each basis order (adding a small percentage to keep safe)
        valence = np.array([50, 200, 400, 800, 1400, 2500])

        # Build nnz pattern for each row
        nnz = np.full((total_num_dofs), valence[model.basis_order-1], dtype=np.int)

        # Build PETSc structures
        vector = createSequentialVectorWithArray(nnz)

        # Build path to save the file
        out_path = output.directory_scratch + '/nnz.dat'

        if parEnv.rank == 0:
            # Write PETGEM nodes in PETSc format
            writePetscVector(out_path, vector, communicator=PETSc.COMM_SELF)

        # ---------------------------------------------------------------
        # Print mesh statistics
        # ---------------------------------------------------------------
        Print.master(' ')
        Print.master('  Mesh statistics')
        Print.master('     Number of elements:   {0:12}'.format(str(nElems)))
        Print.master('     Number of faces:      {0:12}'.format(str(nFaces)))
        Print.master('     Number of edges:      {0:12}'.format(str(nEdges)))
        Print.master('     Number of dofs:       {0:12}'.format(str(total_num_dofs)))
        Print.master('     Number of boundaries: {0:12}'.format(str(len(indx_boundary_dofs))))

        # ---------------------------------------------------------------
        # Print data model
        # ---------------------------------------------------------------
        Print.master(' ')
        Print.master('  Model data')
        Print.master('     Number of materials:    {0:12}'.format(str(np.max(elemsS)+1)))
        Print.master('     Vector basis order:     {0:12}'.format(str(model.basis_order)))
        Print.master('     Frequency (Hz):         {0:12}'.format(str(model.frequency)))
        Print.master('     Source position (xyz):  {0:12}'.format(str(model.src_position)))
        Print.master('     Source azimuth:         {0:12}'.format(str(model.src_azimuth)))
        Print.master('     Source dip:             {0:12}'.format(str(model.src_dip)))
        Print.master('     Source current:         {0:12}'.format(str(model.src_current)))
        Print.master('     Source length:          {0:12}'.format(str(model.src_length)))
        Print.master('     Sigma horizontal:       {0:12}'.format(str(model.sigma_horizontal)))
        Print.master('     Sigma vertical:         {0:12}'.format(str(model.sigma_vertical)))
        Print.master('     Number of receivers:    {0:12}'.format(str(nReceivers)))

        # Apply barrier for MPI tasks alignement
        parEnv.comm.barrier()

        # Stop timer
        Timers()["Preprocessing"].stop()


# ###############################################################
# ################     FUNCTIONS DEFINITION     #################
# ###############################################################
def unitary_test():
    """Unitary test for parallel.py script."""
# ###############################################################
# ################             MAIN             #################
# ###############################################################


if __name__ == '__main__':
    # ---------------------------------------------------------------
    # Run unitary test
    # ---------------------------------------------------------------
    unitary_test()



#!/usr/bin/env python3
# Author:  Octavio Castillo Reyes
# Contact: octavio.castillo@bsc.es
"""Define functions for mesh handling."""

# ---------------------------------------------------------------
# Load python modules
# ---------------------------------------------------------------
import numpy as np
import sys

# ---------------------------------------------------------------
# Load petgem modules (BSC)
# ---------------------------------------------------------------
from .vectors import deleteDuplicateRows


class gmshObject:
    """Class for mesh of type Gmsh.

    This class provides methods to parse .msh files to python format.
    """

    def __init__(self, mshfilename):
        """Init a gmshObject.

        :param str mshFile: mesh file to be initialized.
        :return: gmshObject initialized.
        """
        # Read mesh file
        self.mshfilename = mshfilename
        self.mshfID = open(mshfilename, 'r')

        # Init variables
        readingNodes = 0
        readingElements = 0
        self.nNodes = 0
        self.nElems = 0
        self.physical_groups = []
        self.nodes_in_groups = {}
        self.number_physical_groups = 0
        self.nodes_rules = []
        self.elements_rules = []

        linenumber = 1
        # Read file and identify sections whithin it
        for line in self.mshfID:
            # Get beginning of nodes section
            if line.find('$Nodes') >= 0:
                readingNodes = 1
                continue

            # Get beginning of elements section
            if line.find('$Elements') >= 0:
                readingElements = 1
                continue

            # Get end of elements section
            if line.find('$EndElements') >= 0:
                readingElements = 0
                continue

            # Get end of nodes section
            if line.find('$EndNodes') >= 0:
                readingNodes = 0
                continue

            # Read number of nodes (nNodes) if this
            # is the first line of nodes section
            if readingNodes == 1:
                self.nNodes = np.int32(line)
                readingNodes = 2
                continue

            # Read number of elements (nElems) if this
            # is the first line of elements section
            if readingElements == 1:
                self.nElems = np.int32(line)
                readingElements = 2
                continue

            # Sparse elements and populate the list of nodes in groups
            if readingElements == 2:
                current_line = np.array(line.split(), dtype=np.int32)

                # Get properties from header section
                #eletag = current_line[0]
                #eletype = current_line[1]
                ntags = current_line[2]
                physgrp = 0
                #partition = 0

                if ntags >= 2:
                    physgrp = current_line[3]
                    nodelist = current_line[(3 + ntags)::]

                    if physgrp in self.physical_groups:
                        self.nodes_in_groups[physgrp][nodelist] = 1
                    else:
                        self.nodes_in_groups[physgrp] = -np.ones(self.nNodes+1,
                                                                 dtype=np.int16
                                                                 )
                        self.nodes_in_groups[physgrp][nodelist] = 1
                        self.physical_groups.append(physgrp)

                else:
                    self.__gmshError__(self.mshfilename + '.msh file has < 2 '
                                       'tags at line ' + str(linenumber))

            linenumber += 1

        # Number of physical groups
        self.number_physical_groups = len(self.physical_groups)
        # Close file
        self.mshfID.close()

        return

    def __gmshPrint__(self, msg):
        """Print a message related with a gmshObject.

        :param str msg: message to be printed.
        """
        print(msg)

        return

    def __gmshError__(self, msg):
        """Print error message in a gmshObject and exit.

        :param str msg: message to be printed.
        :return: None.
        """
        sys.stderr.write('   gmshObject: Error! -> ' + msg + '\n')

        return

    def __destroy__(self):
        """Destroy a gmshObject.

        :return: None.
        """
        self.mshfID.close()

        return

    def printNumberNodes(self):
        """Print number of nodes of a gmshObject.

        :return: None.
        """
        print('   Mesh has ' + str(self.nNodes) + ' nodes')

        return

    def printNumberElements(self):
        """Print number of elements of a gmshObject.

        :return: None.
        """
        print('   Mesh has ' + str(self.nElems) + ' elements')

        return

    def printNumberPhysicalGroups(self):
        """Print number of physical groups of a gmshObject.

        :return: None.
        """
        print('   Mesh has ' + str(self.number_physical_groups) +
              ' physical groups')

        return

    def addElementsRule(self, condition, action):
        """Add (append) an user rule to list of elements in the gmshObject.

        :param str condition: function or condition to be added.
        :param str action: action to be executed.
        :return: None.
        """
        self.elements_rules.append((condition, action))


    def addNodesRule(self, condition, action):
        """Add (append) an user rule to list of nodes in the gmshObject.

        :param str condition: function or condition to be added.
        :param str action: action to be executed.
        :return: None.
        """
        self.nodes_rules.append((condition, action))


    def cleanRules(self):
        """Clean rules in the gmshObject."""
        self.nodes_rules = []
        self.elements_rules = []


    def gmshParser(self):
        """Parser nodesRules and nodesElements in an gmshObject."""
        # Open Gmsh file
        self.mshfID = open(self.mshfilename, 'r')

        # Move to nodes section
        line = self.mshfID.readline()
        while(line.find('$Nodes') < 0):
            line = self.mshfID.readline()

        # This line should contain number of nodes (nNodes)
        line = self.mshfID.readline()
        # Verify that nNodes in file is still the nNodes in memory
        if(not np.int32(line) == self.nNodes):
            self.__gmshError__('The number of nodes in ' +
                               self.mshfilename + ' is not consistent.' +
                               ' Aborting')
            exit(-1)

        # Rules for nodes
        if len(self.nodes_rules) == 0:
            # No rules, therefore skip nodes section
            for i in np.arange(self.nNodes):
                self.mshfID.readline()
        else:
            # Read nodes section and apply nodes rules
            for i in np.arange(self.nNodes):
                # Parse the line
                current_line = self.mshfID.readline().split()
                tag = np.int32(current_line[0])
                x = np.double(current_line[1])
                y = np.double(current_line[2])
                z = np.double(current_line[3])

                # Determine the groups to which this node belongs
                physgroups = []
                for grp in self.physical_groups:
                    if self.nodes_in_groups[grp][tag] == 1:
                        physgroups.append(grp)

                for condition, action in self.nodes_rules:
                    if condition(tag, x, y, z, physgroups):
                        action(tag, x, y, z)

        # Read another 2 lines after nodes section is done
        # Read $EndNodes line
        line = self.mshfID.readline()
        # Read #Elements line
        line = self.mshfID.readline()
        # Next line should contain number of elements
        line = self.mshfID.readline()

        # Verify that nElems in file is still the nElems in memory
        if(not np.int32(line) == self.nElems):
            self.__gmshError__('The number of elements in ' +
                               self.mshfilename + '.msh is not consistent.' +
                               ' Aborting')
            exit(-1)

        # Rules for elements
        if len(self.elements_rules) == 0:
            # No rules, therefore skip elements section
            for i in np.arange(self.nElems):
                self.mshfID.readline()
        else:
            # Read elements section and apply nodes rules
            nodes = []
            for i in np.arange(self.nElems):
                # Parse the line
                current_line = self.mshfID.readline().split()
                eletag = np.int32(current_line[0])
                eletype = np.int32(current_line[1])
                ntags = np.int32(current_line[2])
                physgrp = np.int32(current_line[3])
                #partition = np.int32(current_line[4])

                if ntags >= 2:
                    physgrp = np.int32(current_line[3])
                    nodes = np.array(current_line[(3 + ntags)::],
                                     dtype=np.int32)

                    for condition, action in self.elements_rules:
                        if condition(eletag, eletype, physgrp, nodes):
                            action(eletag, eletype, physgrp, nodes)
                else:
                    self.__gmshError__(self.mshfilename + '.msh file has < ' +
                                       '2 tags element with tag ' +
                                       str(eletag))

    # Element definitions in Gmsh
    # 2-node line.
    line_2_node = np.int32(1)
    # 3-node triangle.
    triangle_3_node = np.int32(2)
    # 4-node quadrangle
    quadrangle_4_node = np.int32(3)
    # 4-node tetrahedron.
    tetrahedron_4_node = np.int32(4)
    # 8-node hexahedron.
    hexahedron_8_node = np.int32(5)
    # 6-node prism.
    prism_6_node = np.int32(6)
    # 5-node pyramid.
    pyramid_5_node = np.int32(7)
    # 3-node second order line (2 nodes associated with the vertices and
    # 1 with the edge).
    line_3_node = np.int32(8)
    # 6-node second order triangle (3 nodes associated with the vertices
    # and 3 with the edges).
    triangle_6_node = np.int32(9)
    # 9-node second order quadrangle (4 nodes associated with the vertices,
    # 4 with the edges and 1 with the face).
    quadrangle_9_node = np.int32(10)
    # 10-node second order tetrahedron (4 nodes associated with the vertices
    # and 6 with the edges).
    tetrahedron_10_node = np.int32(11)
    # 27-node second order hexahedron (8 nodes associated with the vertices,
    # 12 with the edges, 6 with the faces and 1 with the volume).
    hexahedron_27_node = np.int32(12)
    # 18-node second order prism (6 nodes associated with the vertices,
    # 9 with the edges and 3 with the quadrangular faces).
    prism_18_node = np.int32(13)
    # 14-node second order pyramid (5 nodes associated with the vertices,
    # 8 with the edges and 1 with the quadrangular face).
    pyramid_14_node = np.int32(14)
    # 1-node point.
    point_1_node = np.int32(15)
    # 8-node second order quadrangle (4 nodes associated with the vertices
    # and 4 with the edges).
    quadrangle_8_node = np.int32(16)
    # 20-node second order hexahedron (8 nodes associated with the vertices
    # and 12 with the edges).
    hexahedron_20_node = np.int32(17)
    # 15-node second order prism (6 nodes associated with the vertices
    # and 9 with the edges).
    prism_15_node = np.int32(18)
    # 13-node second order pyramid (5 nodes associated with the vertices
    # and 8 with the edges).
    pyramid_13_node = np.int32(19)
    # 9-node third order incomplete triangle (3 nodes associated with
    # the vertices, 6 with the edges)
    triangle_9_node_incomplete = np.int32(20)
    # 10-node third order triangle (3 nodes associated with the vertices,
    # 6 with the edges, 1 with the face)
    triangle_10_node = np.int32(21)
    # 12-node fourth order incomplete triangle (3 nodes associated with
    # the vertices, 9 with the edges)
    triangle_12_node_incomplete = np.int32(22)
    # 15-node fourth order triangle (3 nodes associated with the vertices,
    # 9 with the edges, 3 with the face)
    triangle_15_node = np.int32(23)
    # 15-node fifth order incomplete triangle (3 nodes associated with
    # the vertices, 12 with the edges)
    triangle_15_node_incomplete = np.int32(24)
    # 21-node fifth order complete triangle (3 nodes associated with the
    # vertices, 12 with the edges, 6 with the face)
    triangle_21_node = np.int32(25)
    # 4-node third order edge (2 nodes associated with the vertices,
    # 2 internal to the edge)
    edge_4_node = np.int32(26)
    # 5-node fourth order edge (2 nodes associated with the vertices,
    # 3 internal to the edge)
    edge_5_node = np.int32(27)
    # 6-node fifth order edge (2 nodes associated with the vertices,
    # 4 internal to the edge)
    edge_6_node = np.int32(28)
    # 20-node third order tetrahedron (4 nodes associated with the vertices,
    # 12 with the edges, 4 with the faces)
    tetrahedron_20_node = np.int32(29)
    # 35-node fourth order tetrahedron (4 nodes associated with the vertices,
    # 18 with the edges, 12 with the faces, 1 in the volume)
    tetrahedron_35_node = np.int32(30)
    # 56-node fifth order tetrahedron (4 nodes associated with the vertices,
    # 24 with the edges, 24 with the faces, 4 in the volume)
    tetrahedron_56_node = np.int32(31)
    # 64-node third order hexahedron (8 nodes associated with the vertices,
    # 24 with the edges, 24 with the faces, 8 in the volume)
    hexahedron_64_node = np.int32(92)
    # 125-node fourth order hexahedron (8 nodes associated with the vertices,
    # 36 with the edges, 54 with the faces, 27 in the volume)
    hexahedron_125_node = np.int32(93)


def readGmshNodes(mesh_file):
    """Read a mesh nodes from a Gmsh file.

    :param str mesh_file: mesh file to be readed.
    :return: nodes coordinates and number of nodes.
    :rtype: ndarray, int.
    """
    # Create a gmshObject
    inputMesh = gmshObject(mesh_file)

    # Get number of nodes
    nNodes = inputMesh.nNodes

    # Allocate resources
    nodes = np.zeros((nNodes+1, 3), dtype=np.float)

    # Define condition for nodes elemsN, nElems

    def isNode(tag, x, y, z, physgroups):
        """Determine if a point is node.

        :param int tag: node ID.
        :param float x: x coordinate of tag.
        :param float y: y coordinate of tag.
        :param float z: z coordinate of tag.
        :param int physgroups: physgroup of tag.
        """
        return True

    # Define action for nodes
    def getNode(tag, x, y, z):
        """Get coordinates of a node point.

        :param int tag: node ID.
        :param float x: x coordinate of tag.
        :param float y: y coordinate of tag.
        :param float z: z coordinate of tag.
        :return: nodal coordinates.
        :rtype: ndarray
        """
        nodes[tag, :] = [x, y, z]

    # Add nodes rule (composed by a condition and an action)
    inputMesh.addNodesRule(isNode, getNode)

    # Parse mesh file to get nodal coordinates
    inputMesh.gmshParser()

    # Delete first column of nodes since it does not
    # correspond to any node
    nodes = np.delete(nodes, (0), axis=0)

    return nodes, nNodes


def readGmshConnectivity(mesh_file):
    """Read a mesh connectivity from a Gmsh file.

    :param str mesh_file: mesh file to be readed.
    :return: mesh connectivity (elemsN) and number of elements.
    :rtype: ndarray, int.
    """
    # Create a gmshObject
    inputMesh = gmshObject(mesh_file)

    # Get number of nodes
    nElems = inputMesh.nElems

    # Allocate resources
    elemsN = np.zeros((nElems+1, 4), dtype=np.int)

    # Define condition for elements
    def isTetrahedralElement(eletag, eletype, physgrp, nodes):
        """Determine if an element is of linear tetrahedral type.

        :param int eletag: element ID.
        :param int eletype: element type (defined by Gmsh).
        :param int physgrp: physgroup of eletag.
        :param ndarray nodes: nodal indexes of eletag.
        :return: element type
        :rtype: int
        """
        return eletype == inputMesh.tetrahedron_4_node

    # Define action for nodes
    def getElement(eletag, eletype, physgrp, nodes):
        """Get element connectivity.

        :param int eletag: element ID.
        :param int eletype: element type (defined by Gmsh).
        :param int physgrp: physgroup of eletag.
        :param ndarray nodes: nodal indexes of eletag.
        :return: element connectivity
        :rtype: ndarray
        """
        elemsN[eletag, :] = nodes

    # Add nodes rule (composed by a condition and an action)
    inputMesh.addElementsRule(isTetrahedralElement, getElement)

    # Parse mesh file to get nodal coordinates
    inputMesh.gmshParser()

    # Delete first column of connectivity since it does not
    # correspond to any element
    elemsN = np.delete(elemsN, (0), axis=0)
    # Convert to a 0-base numbering in connectivity
    elemsN = elemsN-1

    return elemsN, nElems


def readGmshPhysicalGroups(mesh_file):
    """Read conductivity model from a mesh in Gmsh format.

    :param str mesh_file: mesh file to be readed.
    :return: conductivity model (elemsS) and number of elements.
    :rtype: ndarray, int.
    """
    # Create a gmshObject
    inputMesh = gmshObject(mesh_file)

    # Get number of nodes
    nElems = inputMesh.nElems

    # Allocate resources
    elemsS = np.zeros((nElems+1, 1), dtype=np.int)

    # Define condition for elements
    def isTetrahedralElement(eletag, eletype, physgrp, nodes):
        """Determine if an element is of linear tetrahedral type.

        :param int eletag: element ID.
        :param int eletype: element type (defined by Gmsh).
        :param int physgrp: physgroup of eletag.
        :param ndarray nodes: nodal indexes of eletag.
        :return: element type
        :rtype: int
        """
        return eletype == inputMesh.tetrahedron_4_node

    # Define action for physical groups
    def getPhysicalGroup(eletag, eletype, physgrp, nodes):
        """Get element connectivity.

        :param int eletag: element ID.
        :param int eletype: element type (defined by Gmsh).
        :param int physgrp: physgroup of eletag.
        :param ndarray nodes: nodal indexes of eletag.
        :return: element connectivity
        :rtype: ndarray
        """
        elemsS[eletag] = physgrp

    # Add nodes rule (composed by a condition and an action)
    inputMesh.addElementsRule(isTetrahedralElement, getPhysicalGroup)

    # Parse mesh file to get physical groups for each element
    inputMesh.gmshParser()

    # Delete first column of connectivity since it does not
    # correspond to any element
    elemsS = np.delete(elemsS, (0), axis=0)
    # Convert to a 0-base numbering in the conductivity model
    elemsS = elemsS-1

    return elemsS, nElems


def computeEdges(elemsN, nElems):
    """Compute edges of a 3D tetrahedral mesh.

    :param ndarray elemsN: elements-nodes connectivity.
    :param int nElems: number of tetrahedral elements in the mesh.
    :return: edges connectivity and edgesNodes connectivity.
    :rtype: ndarray
    """
    # Extracts sets of edges
    edges1 = elemsN[:, [0, 1]]
    edges2 = elemsN[:, [1, 2]]
    edges3 = elemsN[:, [0, 2]]
    edges4 = elemsN[:, [0, 3]]
    edges5 = elemsN[:, [1, 3]]
    edges6 = elemsN[:, [2, 3]]

    # Edges as sets of their nodes (vertices)
    vertices = np.zeros([nElems*6, 2])
    vertices[0::6] = edges1
    vertices[1::6] = edges2
    vertices[2::6] = edges3
    vertices[3::6] = edges4
    vertices[4::6] = edges5
    vertices[5::6] = edges6

    # Delete duplicate rows
    [edgesNodes, edges] = deleteDuplicateRows(vertices)

    # Build dofs matrix
    edges = np.array(np.reshape(edges, (nElems, 6)), dtype=np.int)

    # Build dofs to nodes connectivity
    edgesNodes.sort(axis=1)
    edgesNodes = np.array(edgesNodes, dtype=np.int)

    return edges, edgesNodes


def computeFaces(elemsN, nElems):
    r"""Compute the element\'s faces of a 3D tetrahedral mesh.

    :param ndarray matrix: elements-nodes connectivity.
    :param int nElems: number of elements in the mesh.
    :return: element/faces connectivity.
    :rtype: ndarray

    .. note:: References:\n
       Rognes, Marie E., Robert Cndarray. Kirby, and Anders Logg. "Efficient
       assembly of H(div) and H(curl) conforming finite elements."
       SIAM Journal on Scientific Computing 31.6 (2009): 4130-4151.
    """
    # Extracts sets of faces for each nedelec element order
    faces1 = elemsN[:, [0, 1, 2]]
    faces2 = elemsN[:, [0, 1, 3]]
    faces3 = elemsN[:, [1, 2, 3]]
    faces4 = elemsN[:, [0, 2, 3]]

    # Faces as sets of their nodes (vertices)
    vertices = np.zeros([nElems*4, 3])
    vertices[0::4] = faces1
    vertices[1::4] = faces2
    vertices[2::4] = faces3
    vertices[3::4] = faces4

    [facesN, elemsF] = deleteDuplicateRows(vertices)

    numFacesElement = 4
    elemsF = np.array(np.reshape(elemsF, (nElems, numFacesElement)),
                      dtype=np.int)
    facesN = np.array(facesN, dtype=np.int)

    return elemsF, facesN


def computeBoundaryFaces(elemsF, facesN):
    """Compute boundary faces of a tetrahedral mesh.

    :param ndarray elemsF: elements-face connectivity.
    :param ndarray facesN: faces-nodes connectivity.
    :return: nodal-connectivity and indexes of boundary-faces.
    :rtype: ndarray
    """
    # Sort indexes and add 1 position in order to use indexes as Matlab
    A0 = np.sort(elemsF[:, 0]) + 1
    I0 = np.argsort(elemsF[:, 0]) + 1
    A1 = np.sort(elemsF[:, 1]) + 1
    I1 = np.argsort(elemsF[:, 1]) + 1
    A2 = np.sort(elemsF[:, 2]) + 1
    I2 = np.argsort(elemsF[:, 2]) + 1
    A3 = np.sort(elemsF[:, 3]) + 1
    I3 = np.argsort(elemsF[:, 3]) + 1

    # Number of faces
    nFaces = elemsF.max()

    # As consequence, dimensions of E must be increased
    # 2 rows and 1 column
    E = np.zeros((nFaces+2, 9))

    E[A0, 1] = I0
    E[A1, 2] = I1
    E[A2, 3] = I2
    E[A3, 4] = I3

    # If the same face is listed in the same row of 'elemsF'
    # more than, once it will simply be missed! Because of this we
    # have to insert the following dummy variables in order to
    # determine the boundary faces.
    tmp = np.diff(A0) == 0
    ind0 = np.where(tmp)[False]
    tmp = np.diff(A1) == 0
    ind1 = np.where(tmp)[False]
    tmp = np.diff(A2) == 0
    ind2 = np.where(tmp)[False]
    tmp = np.diff(A3) == 0
    ind3 = np.where(tmp)[False]

    E[A0[ind0], 5] = 1
    E[A1[ind1], 6] = 1
    E[A2[ind2], 7] = 1
    E[A3[ind3], 8] = 1

    # Delete extra rows and column
    E = np.delete(E, (0), axis=0)
    E = np.delete(E, (0), axis=1)

    # Final sorting
    E.sort()
    E = np.fliplr(E)

    #  Get boundary nodes by first examining which columns in E
    # have only one nonzero element, meaning that this face is
    # related to only one single tetra, which means it is on the
    # boundary of the domain. Since faces are defined by their nodes,
    # we have the boundary nodes too.
    # Get boundary faces to nodes
    ind = (E[:, 1] == 0)
    bfacesN = np.array(np.transpose(facesN[ind, :]), dtype=np.int)

    # Get indexes of boundary faces
    ind = np.where(ind == True)
    bFaces = np.array(np.transpose(ind), dtype=np.int)
    size = bFaces.shape
    nBoundaryFaces = size[0]
    bFaces = bFaces.reshape((nBoundaryFaces))

    return bfacesN, bFaces


def computeBoundaryEdges(edgesN, bfacesN):
    """Compute boundary edges of a tetrahedral mesh.

    :param ndarray edgesN: edges-nodes connectivity.
    :param ndarray bfacesN: boundary-faces-nodes connectivity.
    :return: boundary-edges connectivity.
    :rtype: ndarray
    """
    # Extracts sets of edges-nodes (add 1 to indexes - Matlab indexing)
    edges1 = (bfacesN[[0, 1], :] + 1).transpose()
    edges2 = (bfacesN[[1, 2], :] + 1).transpose()
    edges3 = (bfacesN[[2, 0], :] + 1).transpose()

    # Number of boundary-faces
    dim = bfacesN.shape
    nBoundaryFaces = dim[1]

    # Boudary faces as sets of their edges (vertices)
    vertices = np.zeros([nBoundaryFaces*3, 2])
    vertices[0::3] = edges1
    vertices[1::3] = edges2
    vertices[2::3] = edges3

    # Repeated setts of nodes (joint edges) are eliminated
    [temp, _] = deleteDuplicateRows(vertices)

    matrixs = np.concatenate((edgesN + 1, temp), axis=0)

    matrixs.sort(axis=1)

    tags = np.lexsort((matrixs[:, 1], matrixs[:, 0]))
    matrixs = matrixs[tags]

    ind0 = np.diff(matrixs[:, 0]) == 0
    ind1 = np.diff(matrixs[:, 1]) == 0

    # Concatenate vectors (vertical stack)
    ind = np.vstack((ind0, ind1))
    ind = ind.transpose()

    # Which ones were reps? k is a vector of indexes to matrix
    k = np.array(np.all(ind, axis=1).ravel().nonzero())

    # tags(k) is an index vector to edgesN (matrix) and denotes those edges
    # which are on boundary tags(k+1) is an index vector to matrix and
    # matrix(tags(k+a)) is the same as bedges, but in different order.
    # I could just return tags(k), but we want that the order is the same
    # as in bEdgesN
    tags2 = np.array(np.argsort(tags[k+1]))

    bEdges = np.array(tags[k[0][tags2]], dtype=np.int)
    bEdges = bEdges[0,:]

    return bEdges



def computeBoundaries(dof_connectivity, dof_edges, dof_faces, bEdges, bFaces, Nord):
    """Compute the indexes of dofs boundaries and internal dofs.

    :param ndarray dof_connectivity: local/global dofs list for elements
    :param ndarray dof_edges: dofs index on edges
    :param ndarray dof_faces: dofs index on faces
    :param ndarray bEdges: boundary-edges connectivity with dimensions = (number_boundary_edges,1)
    :param ndarray bfaces: indexes of boundary-faces = (number_boundary_faces, 1)
    :param int Nord: polynomial order of nedelec basis functions
    :return: indexes of internal dofs and indexes of boundary dofs
    :rtype: ndarray
    """
    # Number of boundaries on edges
    nBoundaryEdges = len(bEdges)
    num_dof_in_edge = Nord

    # Number of boundaries on faces
    nBoundaryFaces = len(bFaces)
    num_dof_in_face = Nord*(Nord-1)

    # Get boundary dofs for edges
    indx_boundary_edges = dof_edges[bEdges,:]

    # Get boundary dofs for faces
    if dof_faces.size == 0:
        # No dofs on faces (first order, Nord==1)
        indx_boundary_faces = np.zeros((1,0), dtype=np.int)
    else:
        indx_boundary_faces = dof_faces[bFaces,:]

    # Get indexes of boundary dofs
    tmp1 = np.reshape(indx_boundary_edges, (nBoundaryEdges*num_dof_in_edge))
    tmp2 = np.reshape(indx_boundary_faces, (nBoundaryFaces*num_dof_in_face))
    indx_boundary_dofs = np.hstack((tmp1, tmp2))

    # Get total number of dofs in the mesh
    total_num_dofs = np.max(dof_connectivity) + 1

    # Get indexes of inner dofs
    indx_inner_dofs = np.setdiff1d(np.arange(0,total_num_dofs), indx_boundary_dofs)

    return indx_inner_dofs, indx_boundary_dofs


def unitary_test():
    """Unitary test for mesh.py script."""


if __name__ == '__main__':
    # Standard module import
    unitary_test()



#!/usr/bin/env python3
# Author:  Octavio Castillo Reyes
# Contact: octavio.castillo@bsc.es
"""Define standard vector and matrix functions."""

# ---------------------------------------------------------------
# Load python modules
# ---------------------------------------------------------------
import numpy as np


# ###############################################################
# ################     FUNCTIONS DEFINITION     #################
# ###############################################################
def deleteDuplicateRows(matrix):
    """Delete duplicate rows in a matrix.

    :param ndarray matrix: input matrix to be processed.
    :return: matrix without duplicate rows
    :rtype: ndarray
    """
    temp = np.copy(matrix)
    temp.sort(axis=1)

    [dummy, J, I] = findUniqueRows(temp, return_index=True,
                                   return_inverse=True)
    out = matrix[J, :]

    return (out, I)


def findUniqueRows(array, return_index=False, return_inverse=False):
    """Find unique rows of a two-dimensional numpy array.

    :param ndarray: array to be processed.
    :param bool return_index: the indices of array that result in the
     unique array.
    :param bool return_inverse: indices of the unique array that
     can be used to reconstruct array.
    :return: unique rows.
    :rtype: ndarray.
    """
    array = np.ascontiguousarray(array)

    # View the rows as a 1D structured array.
    arv = array.view(array.shape[1] * [('', array.dtype)])
    out = np.unique(arv, return_index=return_index,
                    return_inverse=return_inverse)
    if isinstance(out, tuple):
        uarv = out[0]

    else:
        uarv = out

    # Restore the original dimensions.
    uar = uarv.view(array.dtype).reshape((-1, array.shape[1]))

    if isinstance(out, tuple):
        out = (uar,) + out[1:]

    else:
        out = uar

    return out


def is_duplicate_entry(x):
    """Compute number of duplicate entries in a vector.

    :param int-array x: matrix.
    :return: number of duplicate entries.
    :rtype: int.
    """
    counts = np.bincount(x)
    duplicate_entries = np.where(counts > 1)[0]
    num_duplicate = duplicate_entries.size

    return num_duplicate


def invConnectivity(M, nP):
    """Compute the opposite connectivity matrix of M.

    :param ndarray M: connectivity matrix with dimensions = (nElems,eleOrder)
    :param int nP: number of nodes/edges/faces in matrix M.
    :return: connectivity matrix with dimensions = (nNodes,S), (nEdges,S) or (nFaces,S)
    :rtype: ndarray.

    .. note::
        eleOrder determines the number of entities per element in matrix M,
        therefore 4 is the nodal element order, 6 is the edge element order and
        3 is the element order of faces.
        S in the output is the maximum number of elements sharing a given node/edge/face

        If M represents a element/nodes connectivity, the function computes
        a node/elements connectivity.

        If M represents a element/edges connectivity, the function computes
        a edge/elements connectivity.

        If M represents a element/faces connectivity, the function computes
        a faces/elements connectivity.
    """
    # Get matrix dimensions
    size = M.shape

    # Number of items
    nItems = size[0]

    # Element order
    orderEle = size[1]

    # Set maximum valence
    valence = 13

    # Allocate
    N = np.zeros((nP,valence), dtype=np.int)
    nn = np.ones(nP, dtype=np.int)

    # Build matrix
    for i in np.arange(nItems):
        iele = M[i,:]
        nn_Te = nn[iele]
        for j in np.arange(orderEle):
            N[iele[j], nn_Te[j]] = i

        nn[iele] += np.int(1)

    # Delete zeros of matrix N
    N = np.delete(N, np.arange(np.max(nn)-1, valence), axis=1)

    return N


def unitary_test():
    """Unitary test for vectors.py script."""


if __name__ == '__main__':
    unitary_test()



#!/usr/bin/env python3
# Author:  Octavio Castillo Reyes
# Contact: octavio.castillo@bsc.es
"""**PETGEM** kernel for 3D CSEM forward modelling using higg order vector elements."""

if __name__ == '__main__':
    # ---------------------------------------------------------------
    # Load python modules
    # ---------------------------------------------------------------
    import sys
    import petsc4py
    # ---------------------------------------------------------------
    # PETSc init
    # ---------------------------------------------------------------
    petsc4py.init(sys.argv)
    # ---------------------------------------------------------------
    # Load python modules
    # ---------------------------------------------------------------
    # ---------------------------------------------------------------
    # Load petgem modules (BSC)
    # ---------------------------------------------------------------
    from petgem.common import Print, InputParameters, Timers
    from petgem.parallel import MPIEnvironment
    from petgem.preprocessing import Preprocessing
    from petgem.solver import Solver

    # ---------------------------------------------------------------
    # Load system setup (both parameters and dataset configuration)
    # ---------------------------------------------------------------
    # Obtain the MPI environment
    parEnv = MPIEnvironment()

    # Import parameters file
    inputSetup = InputParameters(sys.argv[3], parEnv)

    # Initialize timers
    Timers(inputSetup.output.directory)

    # ---------------------------------------------------------------
    # Print header
    # ---------------------------------------------------------------
    Print.header()

    # ---------------------------------------------------------------
    # Initialize preprocessing and timers
    # ---------------------------------------------------------------
    Print.master(' ')
    Print.master('  Data preprocessing')

    # Create a preprocessing instance and output directory
    preprocessing = Preprocessing()

    # Run preprocessing
    preprocessing.run(inputSetup)

    # ---------------------------------------------------------------
    # Initialize and execute the solver
    # ---------------------------------------------------------------
    Print.master(' ')
    Print.master('  Run modelling')

    # Create a solver instance
    csem_solver = Solver()

    # Setup solver (import files from preprocessing stage)
    csem_solver.setup(inputSetup)

    # Assembly linear system
    csem_solver.assembly(inputSetup)

    # Set dirichlet boundary conditions
    csem_solver.solve()

    # Compute electromagnetic responses
    csem_solver.postprocess(inputSetup)

    # ---------------------------------------------------------------
    # End of PETGEM kernel
    # ---------------------------------------------------------------


















    #     # ###############################################
#     # ------------- Load PETGEM modules -------------
#     from petgem.base.styles import printPetgemHeader
#     from petgem.base.styles import printPetgemFooter
#     from petgem.base.base import readUserParams
#     from petgem.mesh.mesh import getMeshInfo
#     from petgem.efem.efem import defineEFEMConstants
#     from petgem.monitoring.monitoring import printElapsedTimes
#     from petgem.monitoring.monitoring import startLogEvent
#     from petgem.parallel.parallel import printMessage
#     from petgem.parallel.parallel import getRankSize
#     from petgem.parallel.parallel import getRanges
#     from petgem.parallel.parallel import readPetscMatrix
#     from petgem.parallel.parallel import readPetscVector
#     from petgem.parallel.parallel import createParallelMatrix
#     from petgem.parallel.parallel import createParallelVector
#     from petgem.parallel.parallel import parallelAssembler
#     from petgem.solver.solver import setBoundaryConditions
#     from petgem.solver.solver import solveSystem
#     from petgem.postprocessing.postprocessing import postProcessingFields
#
#     # ###############################################
#     # -- Get rank, size and MASTER of the MPI pool --
#     MASTER, rank, size = getRankSize()
#
#     # ###############################################
#     # ----------- Print header (Master) -------------
#     printPetgemHeader(rank)
#
#     # ###############################################
#     # ----------------- User input ------------------
#     printMessage('\nInit', rank)
#     printMessage('='*75, rank)
#     input_params = sys.argv
#
#     # ###############################################
#     # ------- Check and read parameters file --------
#     modelling = readUserParams(input_params, rank)
#
#     # ###############################################
#     # ----- Define geometry and EFEM constants  -----
#     [edgeOrder, nodalOrder,
#      numDimensions] = defineEFEMConstants(modelling['NEDELEC_ORDER'])
#
#     # ###############################################
#     # ---------------- Read mesh --------------------
#     printMessage('\nImport files', rank)
#     printMessage('='*75, rank)
#     # Create and start log event for importing task
#     importLog = startLogEvent("Import_files")
#     importLog.begin()
#     # Read nodes coordinates
#     printMessage('  Nodes coordinates', rank)
#     nodes = readPetscMatrix(modelling['NODES_FILE'], communicator=None)
#     # elements-nodes connectivity
#     printMessage('  Elements-nodes connectivity', rank)
#     elemsN = readPetscMatrix(modelling['MESH_CONNECTIVITY_FILE'],
#                              communicator=None)
#     # elements-faces connectivity
#     printMessage('  Elements-faces connectivity', rank)
#     elemsF = readPetscMatrix(modelling['FACES_CONNECTIVITY_FILE'],
#                              communicator=None)
#     # facesN connectivity
#     printMessage('  Faces-nodes connectivity', rank)
#     facesN = readPetscMatrix(modelling['FACES_NODES_FILE'],
#                              communicator=None)
#     # elements-edges connectivity
#     printMessage('  Elements-edges connectivity', rank)
#     elemsE = readPetscMatrix(modelling['EDGES_CONNECTIVITY_FILE'],
#                              communicator=None)
#     # edgesN connectivity
#     printMessage('  Edges-nodes connectivity', rank)
#     edgesN = readPetscMatrix(modelling['EDGES_NODES_FILE'], communicator=None)
#     # Boundaries
#     printMessage('  Boundary-Edges', rank)
#     boundaries = readPetscVector(modelling['BOUNDARIES_FILE'],
#                                  communicator=None)
#     # Sparsity pattern (NNZ) for matrix allocation
#     printMessage('  Vector for matrix allocation', rank)
#     Q = readPetscVector(modelling['NNZ_FILE'], communicator=None)
#     nnz = (Q.getArray().real).astype(PETSc.IntType)
#     # Conductivity model
#     printMessage('  Conductivity model', rank)
#     elemsSigma = readPetscVector(modelling['CONDUCTIVITY_MODEL_FILE'],
#                                  communicator=None)
#     # Receivers data
#     printMessage('  Receivers data', rank)
#     receivers = readPetscMatrix(modelling['RECEIVERS_FILE'], communicator=None)
#     # End log event for importing task
#     importLog.end()
#
#     # ###############################################
#     # -------------- Mesh information ---------------
#     [nElems, nFaces, nEdges, ndofs,
#      nBoundaries] = getMeshInfo(modelling['NEDELEC_ORDER'], elemsN, elemsF,
#                                 facesN, edgesN, boundaries, rank)
#
#     # ###############################################
#     # --------- Information of parallel pool --------
#     printMessage('\nParallel information', rank)
#     printMessage('='*75, rank)
#     [Istart_elemsE, Iend_elemsE,
#      Istart_boundaries, Iend_boundaries,
#      Istart_receivers, Iend_receivers] = getRanges(elemsE, boundaries,
#                                                    receivers, size, rank)
#
#     # ###############################################
#     # ----- Create and setup parallel structures ----
#     # Left-hand side
#     A = createParallelMatrix(ndofs, ndofs, nnz, modelling['CUDA'],
#                              communicator=None)
#     # Right-hand side
#     b = createParallelVector(ndofs, modelling['CUDA'], communicator=None)
#     # X vector
#     x = createParallelVector(ndofs, modelling['CUDA'], communicator=None)
#
#     # ###############################################
#     # -------------- Parallel assembly --------------
#     printMessage('\nParallel assembly', rank)
#     printMessage('='*75, rank)
#     # Create and start log event for assembly task
#     assemblerLog = startLogEvent("Assembler")
#     assemblerLog.begin()
#     # System assembly
#     [A, b, elapsedTimeAssembly] = parallelAssembler(modelling, A, b, nodes,
#                                                     elemsE, elemsN, elemsF,
#                                                     facesN,  elemsSigma,
#                                                     Istart_elemsE, Iend_elemsE,
#                                                     nEdges, nFaces, rank)
#     # End log event for assembly task
#     assemblerLog.end()
#
#     # ###############################################
#     # ----------- Set boundary conditions -----------
#     printMessage('\nBoundary conditions', rank)
#     printMessage('='*75, rank)
#     # Create and start log event for setting boundary conditions task
#     boundariesLog = startLogEvent("Boundaries")
#     boundariesLog.begin()
#     # Set boundary conditions
#     [A, b, elapsedTimeBoundaries] = setBoundaryConditions(A, b, boundaries,
#                                                           Istart_boundaries,
#                                                           Iend_boundaries,
#                                                           rank)
#     # End log event for setting boundary conditions task
#     boundariesLog.end()
#
#     # ###############################################
#     # ------------------- Solver --------------------
#     printMessage('\nSolver information', rank)
#     printMessage('='*75, rank)
#     # Create and start log event for assembly task
#     solverLog = startLogEvent("Solver")
#     solverLog.begin()
#     # Solve system
#     [x, iterationNumber, elapsedTimeSolver] = solveSystem(A, b, x, rank)
#     # End log event for setting boundary conditions task
#     solverLog.end()
#
#     # ###############################################
#     # --------------- Post-processing ---------------
#     printMessage('\nPost-processing', rank)
#     printMessage('='*75, rank)
#     # # Create and start log event for assembly task
#     postProcessingLog = startLogEvent("Postprocessing")
#     postProcessingLog.begin()
#     elapsedTimepostProcess = postProcessingFields(receivers, modelling, x,
#                                                   Iend_receivers,
#                                                   Istart_receivers,
#                                                   modelling['NEDELEC_ORDER'],
#                                                   modelling['CUDA'],
#                                                   nodalOrder,
#                                                   numDimensions, rank)
#     postProcessingLog.end()
#
#     # ###############################################
#     # -------------- Print elapsed times-------------
#     printMessage('\nElapsed times (seconds)', rank)
#     printMessage('='*75, rank)
#     printElapsedTimes(elapsedTimeAssembly, elapsedTimeSolver,
#                       elapsedTimepostProcess, iterationNumber, rank)
#
#     # ###############################################
#     # ----------- Print footer (Master) -------------
#     printPetgemFooter(rank)
# else:
#     pass



#!/usr/bin/env python3
# Author:  Octavio Castillo Reyes
# Contact: octavio.castillo@bsc.es
"""Define functions a 3D CSEM solver using high-order vector finite element method (HEFEM)."""

# ---------------------------------------------------------------
# Load python modules
# ---------------------------------------------------------------
import numpy as np
from petsc4py import PETSc
import h5py
import shutil

# ---------------------------------------------------------------
# Load petgem modules (BSC)
# ---------------------------------------------------------------
from .common import Print, Timers, measure_all_class_methods
from .parallel import readPetscMatrix, readPetscVector, createParallelMatrix, createParallelVector
from .parallel import MPIEnvironment, createSequentialVector
from .parallel import writePetscVector
from .hvfem import computeJacobian, computeElementOrientation, computeElementalMatrices, computeSourceVectorRotation
from .hvfem import tetrahedronXYZToXiEtaZeta, computeBasisFunctions

# ###############################################################
# ################     CLASSES DEFINITION      ##################
# ###############################################################
@measure_all_class_methods
class Solver():
    """Class for solver."""

    def __init__(self):
        """Initialization of a solver class."""
        return


    def setup(self, inputSetup):
        """Setup of a solver class.

        :param object inputSetup: user input setup.
        """
        # ---------------------------------------------------------------
        # Initialization
        # ---------------------------------------------------------------
        # Start timer
        Timers()["Setup"].start()

        # Parameters shortcut (for code legibility)
        output = inputSetup.output

        Print.master('     Importing files')

        # ---------------------------------------------------------------
        # Import files
        # ---------------------------------------------------------------
        # Read nodes coordinates
        input_file = output.directory_scratch + '/nodes.dat'
        self.nodes = readPetscMatrix(input_file, communicator=None)

        # elements-nodes connectivity
        input_file = output.directory_scratch + '/meshConnectivity.dat'
        self.elemsN = readPetscMatrix(input_file, communicator=None)

        # elements-edges connectivity
        input_file = output.directory_scratch + '/edges.dat'
        self.elemsE = readPetscMatrix(input_file, communicator=None)

        # edges-nodes connectivity
        input_file = output.directory_scratch + '/edgesNodes.dat'
        self.edgesNodes = readPetscMatrix(input_file, communicator=None)

        # elements-faces connectivity
        input_file = output.directory_scratch + '/faces.dat'
        self.elemsF = readPetscMatrix(input_file, communicator=None)

        # faces-edges connectivity
        input_file = output.directory_scratch + '/facesEdges.dat'
        self.facesEdges = readPetscMatrix(input_file, communicator=None)

        # Dofs connectivity
        input_file = output.directory_scratch + '/dofs.dat'
        self.dofs = readPetscMatrix(input_file, communicator=None)

        # Boundaries
        input_file = output.directory_scratch + '/boundaries.dat'
        self.boundaries = readPetscVector(input_file, communicator=None)

        # Conductivity model
        input_file = output.directory_scratch + '/conductivityModel.dat'
        self.sigmaModel = readPetscMatrix(input_file, communicator=None)

        # Receivers
        input_file = output.directory_scratch + '/receivers.dat'
        self.receivers = readPetscMatrix(input_file, communicator=None)

        # Sparsity pattern (NNZ) for matrix allocation
        input_file = output.directory_scratch + '/nnz.dat'
        tmp = readPetscVector(input_file, communicator=None)
        self.nnz = (tmp.getArray().real).astype(PETSc.IntType)

        # Number of dofs (length of nnz correspond to the total number of dofs)
        self.total_num_dofs = tmp.getSizes()[1]     # Get global sizes

        # Stop timer
        Timers()["Setup"].stop()

        return


    def assembly(self, inputSetup):
        """Assembly a linear system for 3D CSEM based on HEFEM.

        :param object inputSetup: user input setup.
        """
        # ---------------------------------------------------------------
        # Initialization
        # ---------------------------------------------------------------
        # Start timer
        Timers()["Assembly"].start()

        # Parameters shortcut (for code legibility)
        model = inputSetup.model
        run = inputSetup.run
        output = inputSetup.output

        Print.master('     Assembling linear system')

        # ---------------------------------------------------------------
        # Get ranges
        # ---------------------------------------------------------------
        # Ranges over elements
        Istart_elemsE, Iend_elemsE = self.elemsE.getOwnershipRange()

        # ---------------------------------------------------------------
        # Allocate parallel arrays
        # ---------------------------------------------------------------
        # Left-hand side
        self.A = createParallelMatrix(self.total_num_dofs, self.total_num_dofs, self.nnz, run.cuda, communicator=None)
        # Right-hand side
        self.b = createParallelVector(self.total_num_dofs, run.cuda, communicator=None)
        # X vector
        self.x = createParallelVector(self.total_num_dofs, run.cuda, communicator=None)

        # ---------------------------------------------------------------
        # Assembly linear system (Left-Hand Side - LHS)
        # ---------------------------------------------------------------
        num_nodes_per_element = 4
        num_edges_per_element = 6
        num_faces_per_element = 4
        #num_nodes_per_face    = 3
        num_edges_per_face    = 3
        num_nodes_per_edge    = 2
        num_dimensions        = 3
        omega                 = model.frequency*2.*np.pi
        mu                    = 4.*np.pi*1e-7
        Const                 = np.sqrt(-1. + 0.j)*omega*mu

        # Compute contributions for all local elements
        for i in np.arange(Istart_elemsE, Iend_elemsE):
            # Get indexes of nodes for i
            nodesEle = (self.elemsN.getRow(i)[1].real).astype(PETSc.IntType)

            # Get coordinates of i
            coordEle = self.nodes.getRow(i)[1].real
            coordEle = np.reshape(coordEle, (num_nodes_per_element, num_dimensions))

            # Get edges indexes for faces in i
            edgesFace = self.facesEdges.getRow(i)[1].real
            edgesFace = np.reshape(edgesFace, (num_faces_per_element, num_edges_per_face))

            # Get indexes of edges for i
            edgesEle = (self.elemsE.getRow(i)[1].real).astype(PETSc.IntType)

            # Get node indexes for edges in i
            edgesNodesEle = self.edgesNodes.getRow(i)[1].real
            edgesNodesEle = np.reshape(edgesNodesEle, (num_edges_per_element, num_nodes_per_edge))

            # Get conductivity values for i (horizontal and vertical conductivity)
            sigmaEle = self.sigmaModel.getRow(i)[1].real

            # Compute jacobian for i
            jacobian, invjacobian = computeJacobian(coordEle)

            # Compute global orientation for i
            edge_orientation, face_orientation = computeElementOrientation(edgesEle,nodesEle,edgesNodesEle,edgesFace)

            # Compute elemental matrices (stiffness and mass matrices)
            M, K = computeElementalMatrices(edge_orientation, face_orientation, jacobian, invjacobian, model.basis_order, sigmaEle)

            # Compute elemental matrix
            Ae = K - Const*M
            Ae = Ae.flatten()

            # Get dofs indexes for i
            dofsEle = (self.dofs.getRow(i)[1].real).astype(PETSc.IntType)

            # Add local contributions to global matrix
            self.A.setValues(dofsEle, dofsEle, Ae, addv=PETSc.InsertMode.ADD_VALUES)


        # Start global LHS assembly
        self.A.assemblyBegin()
        # End global LHS assembly
        self.A.assemblyEnd()

        # ---------------------------------------------------------------
        # Assembly linear system (Right-Hand Side RHS)
        # ---------------------------------------------------------------
        # Compute matrices for source rotation
        sourceRotationVector = computeSourceVectorRotation(model)

        # Total electric field formulation. Set dipole definition
        # x-directed dipole
        Dx = np.array([model.src_current*model.src_length*1., 0., 0.], dtype=np.float)
        # y-directed dipole
        Dy = np.array([0., model.src_current*model.src_length*1., 0.], dtype=np.float)
        # % z-directed dipole
        Dz = np.array([0., 0., model.src_current*model.src_length*1.], dtype=np.float)

        # Rotate source and setup electric field
        field = sourceRotationVector[0]*Dx + sourceRotationVector[1]*Dy + sourceRotationVector[2]*Dz

        # Add source (only master)
        if MPIEnvironment().rank == 0:
            # Read source file
            input_file = output.directory_scratch + '/source.dat'
            source_data = readPetscVector(input_file, communicator=PETSc.COMM_SELF)

            # Get source data
            source_data = source_data.getArray().real

            # Get indexes of nodes for srcElem
            nodesEle = source_data[0:4].astype(np.int)

            # Get nodes coordinates for srcElem
            coordEle = source_data[4:16]
            coordEle = np.reshape(coordEle, (num_nodes_per_element, num_dimensions))

            # Get faces indexes for srcElem
            #facesEle = source_data[16:20].astype(np.int)

            # Get edges indexes for faces in srcElem
            edgesFace = source_data[20:32].astype(np.int)
            edgesFace = np.reshape(edgesFace, (num_faces_per_element, num_edges_per_face))

            # Get indexes of edges for srcElem
            edgesEle = source_data[32:38].astype(np.int)

            # Get node indexes for edges in srcElem
            edgesNodesEle = source_data[38:50].astype(np.int)
            edgesNodesEle = np.reshape(edgesNodesEle, (num_edges_per_element, num_nodes_per_edge))

            # Get dofs for srcElem
            dofsSource = source_data[50::].astype(PETSc.IntType)

            # Compute jacobian for srcElem
            _, invjacobian = computeJacobian(coordEle)

            # Compute global orientation for srcElem
            edge_orientation, face_orientation = computeElementOrientation(edgesEle,nodesEle,edgesNodesEle,edgesFace)

            # Transform xyz source position to XiEtaZeta coordinates (reference tetrahedral element)
            XiEtaZeta = tetrahedronXYZToXiEtaZeta(coordEle, model.src_position)

            # Compute basis for srcElem
            basis = computeBasisFunctions(edge_orientation, face_orientation, invjacobian, model.basis_order, XiEtaZeta)

            # Compute integral
            src_contribution = np.matmul(field, basis[:,:,0])

            # Multiplication by constant value
            src_contribution = src_contribution * Const

            # Add local contributions to global matrix
            self.b.setValues(dofsSource, src_contribution, addv=PETSc.InsertMode.ADD_VALUES)

        # Start global RHS assembly
        self.b.assemblyBegin()
        # End global RHS assembly
        self.b.assemblyEnd()

        # Stop timer
        Timers()["Assembly"].stop()

        return


    def solve(self):
        """Solver for a linear system generated by the HEFEM for a 3D CSEM problem."""
        # ---------------------------------------------------------------
        # Initialization
        # ---------------------------------------------------------------
        # Start timer
        Timers()["SetBoundaries"].start()

        Print.master('     Solving linear system')

        # ---------------------------------------------------------------
        # Set dirichlet boundary conditions
        # ---------------------------------------------------------------
        # Ranges over boundaries
        Istart_boundaries, Iend_boundaries = self.boundaries.getOwnershipRange()
        # Boundaries for LHS
        self.A.zeroRowsColumns(np.real(self.boundaries).astype(PETSc.IntType))
        # Boundaries for RHS
        numLocalBoundaries = Iend_boundaries - Istart_boundaries
        self.b.setValues(np.real(self.boundaries).astype(PETSc.IntType),
                         np.zeros(numLocalBoundaries, dtype=np.complex),
                         addv=PETSc.InsertMode.INSERT_VALUES)

        # Start global system assembly
        self.A.assemblyBegin()
        self.b.assemblyBegin()
        # End global system assembly
        self.A.assemblyEnd()
        self.b.assemblyEnd()

        # Stop timer
        Timers()["SetBoundaries"].stop()

        # ---------------------------------------------------------------
        # Solve system
        # ---------------------------------------------------------------
        Timers()["Solver"].start()
        # Create KSP: linear equation solver
        ksp = PETSc.KSP().create(comm=PETSc.COMM_WORLD)
        ksp.setOperators(self.A)
        ksp.setFromOptions()
        ksp.solve(self.b, self.x)
        iterationNumber = ksp.getIterationNumber()
        ksp.destroy()
        Print.master('     Number of solver iterations: ' + str(iterationNumber))
        Timers()["Solver"].stop()

        return


    def postprocess(self, inputSetup):
        """Interpolate a given electric field for a vector of receivers.

        :param object inputSetup: user input setup.
        """
        # ---------------------------------------------------------------
        # Initialization
        # ---------------------------------------------------------------
        # Start timer
        Timers()["Postprocessing"].start()

        Print.master('     Postprocessing solution')

        # Parameters shortcut (for code legibility)
        model = inputSetup.model
        run = inputSetup.run
        output = inputSetup.output

        # Ranges over receivers
        Istart_receivers, Iend_receivers = self.receivers.getOwnershipRange()

        # Number of receivers
        total_num_receivers = self.receivers.getSize()[0]
        local_num_receivers = Iend_receivers-Istart_receivers

        # Define constants
        num_nodes_per_element = 4
        num_faces_per_element = 4
        num_edges_per_face    = 3
        num_edges_per_element = 6
        num_nodes_per_edge    = 2
        num_dimensions        = 3

        # Compute number of dofs per element
        num_dof_in_element = np.int(model.basis_order*(model.basis_order+2)*(model.basis_order+3)/2)

        # ---------------------------------------------------------------
        # Get dofs-connectivity for receivers
        # ---------------------------------------------------------------
        # Auxiliar arrays
        dofsIdxRecv = np.zeros((local_num_receivers, num_dof_in_element), dtype=PETSc.IntType)

        j = 0
        for i in np.arange(Istart_receivers, Iend_receivers):
            # Get dofs for receiver
            tmp = (self.receivers.getRow(i)[1].real).astype(PETSc.IntType)
            tmp = tmp[53::]
            dofsIdxRecv[j, :] = tmp
            j += 1

        # Gather global solution of x to local vector
        # Sequential vector for gather tasks
        x_local = createSequentialVector(local_num_receivers*num_dof_in_element, run.cuda, communicator=None)

        # Build Index set in PETSc format
        IS_dofs = PETSc.IS().createGeneral(dofsIdxRecv.flatten(), comm=PETSc.COMM_WORLD)

        # Build gather vector
        gatherVector = PETSc.Scatter().create(self.x, IS_dofs, x_local, None)
        # Ghater values
        gatherVector.scatter(self.x, x_local, PETSc.InsertMode.INSERT_VALUES, PETSc.ScatterMode.FORWARD)

        # ---------------------------------------------------------------
        # Allocate parallel arrays for output
        # ---------------------------------------------------------------
        receiver_fieldX = createParallelVector(total_num_receivers, run.cuda ,communicator=None)
        receiver_fieldY = createParallelVector(total_num_receivers, run.cuda ,communicator=None)
        receiver_fieldZ = createParallelVector(total_num_receivers, run.cuda ,communicator=None)
        receiver_coordinatesX = createParallelVector(total_num_receivers, run.cuda ,communicator=None)
        receiver_coordinatesY = createParallelVector(total_num_receivers, run.cuda ,communicator=None)
        receiver_coordinatesZ = createParallelVector(total_num_receivers, run.cuda ,communicator=None)

        k = 0
        for i in np.arange(Istart_receivers, Iend_receivers):
            # Get receiver data
            receiver_data = self.receivers.getRow(i)[1].real

            # Get indexes of nodes for i
            nodesEle = receiver_data[0:4].astype(np.int)

            # Get nodes coordinates for i
            coordEle = receiver_data[4:16]
            coordEle = np.reshape(coordEle, (num_nodes_per_element, num_dimensions))

            # Get faces indexes for i
            #facesEle = receiver_data[16:20].astype(np.int)

            # Get edges indexes for faces in i
            edgesFace = receiver_data[20:32].astype(np.int)
            edgesFace = np.reshape(edgesFace, (num_faces_per_element, num_edges_per_face))

            # Get indexes of edges for i
            edgesEle = receiver_data[32:38].astype(np.int)

            # Get node indexes for edges in i
            edgesNodesEle = receiver_data[38:50].astype(np.int)
            edgesNodesEle = np.reshape(edgesNodesEle, (num_edges_per_element, num_nodes_per_edge))

            # Get receiver coordinates
            coordReceiver = receiver_data[50:53]

            # Get dofs for i
            #dofsSource = receiver_data[53::].astype(PETSc.IntType)

            # Compute jacobian for i
            _, invjacobian = computeJacobian(coordEle)

            # Compute global orientation for i
            edge_orientation, face_orientation = computeElementOrientation(edgesEle,nodesEle,edgesNodesEle,edgesFace)

            # Transform xyz source position to XiEtaZeta coordinates (reference tetrahedral element)
            XiEtaZeta = tetrahedronXYZToXiEtaZeta(coordEle, coordReceiver)

            # Compute basis for i
            basis = computeBasisFunctions(edge_orientation, face_orientation, invjacobian, model.basis_order, XiEtaZeta)

            # Get element fields
            local_field = x_local[k*num_dof_in_element : (k * num_dof_in_element) + num_dof_in_element]

            # Initialize variables
            Ex = 0.
            Ey = 0.
            Ez = 0.

            # Interpolate electric field
            for j in np.arange(num_dof_in_element):
                Rfield = np.real(local_field[j]) # Real part
                Ifield = np.imag(local_field[j]) # Imaginary part
                # Exyz[i] = Exyz[i] + real_part*basis + imag_part*basis
                Ex += Rfield*basis[0,j] + np.sqrt(-1. + 0.j)*Ifield*basis[0,j]
                Ey += Rfield*basis[1,j] + np.sqrt(-1. + 0.j)*Ifield*basis[1,j]
                Ez += Rfield*basis[2,j] + np.sqrt(-1. + 0.j)*Ifield*basis[2,j]

            # Increase counter over local vector (x_local)
            k += 1

            # Set total field components for i
            receiver_fieldX.setValue(i, Ex, addv=PETSc.InsertMode.INSERT_VALUES)
            receiver_fieldY.setValue(i, Ey, addv=PETSc.InsertMode.INSERT_VALUES)
            receiver_fieldZ.setValue(i, Ez, addv=PETSc.InsertMode.INSERT_VALUES)

            # Set coordinates for i
            receiver_coordinatesX.setValue(i, coordReceiver[0], addv=PETSc.InsertMode.INSERT_VALUES)
            receiver_coordinatesY.setValue(i, coordReceiver[1], addv=PETSc.InsertMode.INSERT_VALUES)
            receiver_coordinatesZ.setValue(i, coordReceiver[2], addv=PETSc.InsertMode.INSERT_VALUES)

        # Start global assembly
        receiver_fieldX.assemblyBegin()
        receiver_fieldY.assemblyBegin()
        receiver_fieldZ.assemblyBegin()
        receiver_coordinatesX.assemblyBegin()
        receiver_coordinatesY.assemblyBegin()
        receiver_coordinatesZ.assemblyBegin()

        # End global assembly
        receiver_fieldX.assemblyEnd()
        receiver_fieldY.assemblyEnd()
        receiver_fieldZ.assemblyEnd()
        receiver_coordinatesX.assemblyEnd()
        receiver_coordinatesY.assemblyEnd()
        receiver_coordinatesZ.assemblyEnd()

        # Write intermediate results
        out_path = output.directory_scratch + '/fieldsX.dat'
        writePetscVector(out_path, receiver_fieldX, communicator=None)

        out_path = output.directory_scratch + '/fieldsY.dat'
        writePetscVector(out_path, receiver_fieldY, communicator=None)

        out_path = output.directory_scratch + '/fieldsZ.dat'
        writePetscVector(out_path, receiver_fieldZ, communicator=None)

        out_path = output.directory_scratch + '/receiver_coordinatesX.dat'
        writePetscVector(out_path, receiver_coordinatesX, communicator=None)

        out_path = output.directory_scratch + '/receiver_coordinatesY.dat'
        writePetscVector(out_path, receiver_coordinatesY, communicator=None)

        out_path = output.directory_scratch + '/receiver_coordinatesZ.dat'
        writePetscVector(out_path, receiver_coordinatesZ, communicator=None)

        # Write final solution
        if MPIEnvironment().rank == 0:
            input_file = output.directory_scratch + '/fieldsX.dat'
            electric_fieldsX = readPetscVector(input_file, communicator=PETSc.COMM_SELF)

            input_file = output.directory_scratch + '/fieldsY.dat'
            electric_fieldsY = readPetscVector(input_file, communicator=PETSc.COMM_SELF)

            input_file = output.directory_scratch + '/fieldsZ.dat'
            electric_fieldsZ = readPetscVector(input_file, communicator=PETSc.COMM_SELF)

            input_file = output.directory_scratch + '/receiver_coordinatesX.dat'
            recv_coordX = readPetscVector(input_file, communicator=PETSc.COMM_SELF)

            input_file = output.directory_scratch + '/receiver_coordinatesY.dat'
            recv_coordY = readPetscVector(input_file, communicator=PETSc.COMM_SELF)

            input_file = output.directory_scratch + '/receiver_coordinatesZ.dat'
            recv_coordZ = readPetscVector(input_file, communicator=PETSc.COMM_SELF)

            # Allocate
            data_fields = np.zeros((total_num_receivers, 3), dtype=np.complex)
            data_coordinates = np.zeros((total_num_receivers, 3), dtype=np.float)

            # Loop over receivers
            for i in np.arange(total_num_receivers):
                # Get electric field components
                data_fields[i, 0] = electric_fieldsX.getValue(i)
                data_fields[i, 1] = electric_fieldsY.getValue(i)
                data_fields[i, 2] = electric_fieldsZ.getValue(i)

                # Get coordinates
                data_coordinates[i, 0] = recv_coordX.getValue(i).real
                data_coordinates[i, 1] = recv_coordY.getValue(i).real
                data_coordinates[i, 2] = recv_coordZ.getValue(i).real

            # Write final output
            output_file = output.directory + '/electric_fields.h5'
            fileID = h5py.File(output_file, 'w')

            # Create coordinates dataset
            _ = fileID.create_dataset('electric_fields', data=data_fields)
            _ = fileID.create_dataset('receiver_coordinates', data=data_coordinates)

            # Close file
            fileID.close()

            # Remove temporal directory
            shutil.rmtree(output.directory_scratch)

        # Stop timer
        Timers()["Postprocessing"].stop()


def unitary_test():
    """Unitary test for hvfem.py script."""


if __name__ == '__main__':
    unitary_test()



#!/usr/bin/env python3
# Author:  Octavio Castillo Reyes
# Contact: octavio.castillo@bsc.es
"""Define common operations for **PETGEM**."""

# ---------------------------------------------------------------
# Load python modules
# ---------------------------------------------------------------
import time
import yaml
import sys
import os
import numpy as np
from functools import wraps
from colorama import Fore
from singleton_decorator import singleton

# ---------------------------------------------------------------
# Load petgem modules (BSC)
# ---------------------------------------------------------------
from .parallel import MPIEnvironment

# ###############################################################
# ################     CLASSES DEFINITION      ##################
# ###############################################################

# ---------------------------------------------------------------
# Class Print definition
# ---------------------------------------------------------------
class Print(object):
    """
    This class provides methods for pretty print.

    :param object str: string to be printed.
    :return: None.
    :rtype: None.

    """
    # Options for Gauss points computation (switch case)
    _options = {
        1: Fore.BLACK,
        2: Fore.BLUE,
        3: Fore.CYAN,
        4: Fore.GREEN,
        5: Fore.LIGHTBLACK_EX,
        6: Fore.LIGHTBLUE_EX,
        7: Fore.LIGHTCYAN_EX,
        8: Fore.LIGHTGREEN_EX,
        9: Fore.LIGHTMAGENTA_EX,
        10: Fore.LIGHTRED_EX,
        11: Fore.LIGHTWHITE_EX,
        12: Fore.LIGHTYELLOW_EX,
        13: Fore.MAGENTA,
        14: Fore.RED,
        15: Fore.WHITE,
        16: Fore.YELLOW
    }

    # Constructor
    def __init__(self, text, color_code=None):
        """Constructor."""
        self._log(text, color_code)

    # Logging method
    def _log(self, text, color_code=None):
        """
        Configure and prints a text.

        :param str text: text to be printed.
        :param int color_code: text color code.
        :return: None.

        """
        # Verify if color_code is None, then use black color
        if color_code is None:
            color_code = int(16)

        set_color = self._options[color_code]
        print(set_color + text)
        sys.stdout.flush()
        return

    @classmethod
    def header(self):
        """Print the header.

        :param: None.
        :return: None.
        :rtype: None.
        """
        # Specific color code for printing the header
        color_code = 5

        if( MPIEnvironment().rank == 0 ):
            self._log(self, '%'*75, color_code)
            self._log(self, '%%%' + ' '*69 + '%%%', color_code)
            self._log(self, '%%%'+  'PETGEM'.center(69) + '%%%', color_code)
            self._log(self, '%%%'+  'Parallel Edge-based Tool for Electromagnetic Modelling'.center(69) + '%%%', color_code)
            self._log(self, '%%%' + ' '*69 + '%%%', color_code)
            self._log(self, '%'*75, color_code)
            self._log(self, '%%%' + ' '*69 + '%%%', color_code)
            self._log(self, '%%%   (c) Octavio Castillo-Reyes' +
                       ' '*40 + '%%%', color_code)
            self._log(self, '%%%   Barcelona Supercomputing Center, 2020' +
                       ' '*29 + '%%%', color_code)
            self._log(self, '%%%' + ' '*69 + '%%%', color_code)
            self._log(self, '%'*75, color_code)

        return

    @classmethod
    def master(self, text, color_code=None):
        """
        If the caller is the master process, this method prints a message.

        :param: None.
        :return: None.
        :rtype: None.

        """
        if( MPIEnvironment().rank == 0 ):
            self._log(self, text, color_code)

        return

# ---------------------------------------------------------------
# Class InputParameters definition
# ---------------------------------------------------------------
class InputParameters(object):
    """
    Method import a yaml parameter file.

    :param dict object: user params yaml file.
    :return: user parameters as object view.
    :rtype: object.
    """

    def __init__(self, params, parEnv):
        """Class constructor.

        :param str params: yaml parameters file.
        :param object parEnv: parallel environment object.
        :return: InputParameters object.
        :rtype: object
        """
        # ---------------------------------------------------------------
        # Read the input parameters file
        # ---------------------------------------------------------------
        with open(params, 'r') as f:
            # The FullLoader parameter handles the conversion from YAML
            # scalar values to Python the dictionary format
            inputs = yaml.safe_load(f)

        # Store the datasets as a dict  object
        self.model = Dictionary2Object(inputs['model'])
        self.model.src_position = np.asarray(self.model.src_position, dtype=np.float)
        self.run = Dictionary2Object(inputs['run'])
        self.output = Dictionary2Object(inputs['output'])

        # ---------------------------------------------------------------
        # Validations
        # ---------------------------------------------------------------
        if self.model.basis_order < 1 or self.model.basis_order > 6:
            Print.master('     Vector finite element basis order not supported')
            exit(-1)

        #input(type(self.run.cuda))

        if self.run.cuda is False:
            self.run.cuda = False
        elif self.run.cuda is True:
            self.run.cuda = True
        else:
            Print.master('     Cuda option not supported')
            exit(-1)

        # ---------------------------------------------------------------
        # Create output directory
        # ---------------------------------------------------------------
        if(parEnv.rank == 0):
            if not os.path.exists(self.output.directory):
                os.mkdir(self.output.directory)

            # Create temporal directory
            if not os.path.exists(self.output.directory_scratch):
                os.mkdir(self.output.directory_scratch)

        return


# ---------------------------------------------------------------
# Class dictionary to object definition
# ---------------------------------------------------------------
class Dictionary2Object(object):
    """Turns a dictionary into a class."""

    def __init__(self, dictionary):
        """Constructor."""
        for key in dictionary:
            setattr(self, key, dictionary[key])


# ---------------------------------------------------------------
# Class Timer definition
# ---------------------------------------------------------------
class Timer():
    """
    Definition of timer class.

    """

    # Attributes
    elapsed = None

    # Initializing
    def __init__(self, elapsed = 0):
        """Constructor."""
        self._start = 0
        self.elapsed = elapsed

    # Start timer
    def start(self):
        """Start timer."""
        self._start = time.time()

    # Stop timer
    def stop(self):
        """Stop timer."""
        if(self._start > 0):
            self.elapsed += time.time() - self._start
            self._start = 0

    # Reset timer
    def reset(self):
        """Reset timer."""
        self.elapsed = 0

# ---------------------------------------------------------------
# Class Timers definition
# ---------------------------------------------------------------
@singleton
class Timers():
    """
    Defintion of timers class.

    """

    # Initializing
    def __init__(self, opath = None):
        """Constructor."""
        # Create private dict for storing the timers
        self._elems = {}

        # Set the output path for writting the times report
        self._opath = opath

        # Obtain the process rank once
        self.rank = MPIEnvironment().rank

        # Prepare the output file
        self._out = None
        self._header = False
        if  MPIEnvironment().rank == 0 and self._opath:
            # Save times for parallel scalability test
            out_filename = self._opath + '/times.txt'
            if os.path.isfile(out_filename):
                self._out = open(out_filename, 'a')
            else:
                self._out = open(out_filename, 'w')
                self._header = True

    # Get an specific timer
    def __getitem__(self, key):
        """Get item name for timer."""
        return self._elems.setdefault(key, Timer())

    # Set a specific
    def __setitem__(self, key, value):
        """Set item name for timer."""
        self._elems[key] = Timer(value)

    # Write report
    def _write(self):
        """Write timer."""
        # Obtain the MPI environment
        parEnv = MPIEnvironment()

        # Create empty lists
        block = []
        time = []

        for key, value in self._elems.items():
            total = parEnv.comm.allreduce(value.elapsed, op=parEnv.MPI.SUM) / parEnv.num_proc
            block.append(key)
            time.append(total)

        if( parEnv.rank == 0):
            if( self._header):
                self._out.write("  |  # proc  |" + "".join('%20s  | ' % t for t in block) + "\n")
            self._out.write("  |  %-6s  |" % parEnv.num_proc + "".join('%20.3f  | ' % t for t in time) + "\n")
            self._out.close()

    # Deleting
    def __del__(self):
        """Delete timers."""
        rank = MPIEnvironment().rank
        # For each stored element
        for key, value in self._elems.items():
            msg = ' [Process {:2d}] | {:40s} | {:4.3f} seconds'.format(rank, key, value.elapsed)
            Print(msg, color_code=4)

        self._write()

# ###############################################################
# ################     DECORATORS DEFINITION    #################
# ###############################################################

# ---------------------------------------------------------------
# Decorators for code instrumentation
# ---------------------------------------------------------------
def measure_time(f = None, group = None, split = False):
    """"Implement method to measure execution time.

    Args:
        f: the decorated function
        group: the group name
        split: decides if all blocks in a group contribute to the same timer
    Returns:
        a function wrap

    """
    def inner_function(function):
        @wraps(function)
        def wrapper(*args, **kwargs):

            # No group was specified
            if not group:
                name = function.__name__
            # Each method is measured separately
            elif split:
                name = group + "." + function.__name__
            # Each method contribute to the group time measure
            else:
                name = group

            # Timer start
            Timers()[name].start()

            # Execute the decorated method
            ret = function(*args, **kwargs)

            # Timer stop
            Timers()[name].stop()

            return ret
        return wrapper

    # For the decorate a method without providing arguments
    if f:
        return inner_function(f)
    return inner_function


def measure_all_class_methods(Cls):
    """"Implement a decorator to measure execution time for each method.

    Args:
        f: the decorated function
    Returns:
        a function wrap

    """

    class DecoratedClass(object):

        def __init__(self,*args,**kwargs):
            self.oInstance = Cls(*args,**kwargs)
            self.className = self.oInstance.__class__.__name__

        def __getattribute__(self,s):
            try:
                x = super(DecoratedClass, self).__getattribute__(s)
            except AttributeError:
                pass
            else:
                return x
            x = self.oInstance.__getattribute__(s)
            if type(x) == type(self.__init__): # it is an instance method
                return measure_time(x, group=self.className, split=True)
            else:
                return x
    return DecoratedClass

# ###############################################################
# ################     FUNCTIONS DEFINITION     #################
# ###############################################################

def unitary_test():
    """Unitary test for common.py script."""
    # TODO


# ###############################################################
# ################             MAIN             #################
# ###############################################################
if __name__ == '__main__':
    # ---------------------------------------------------------------
    # Run unitary test
    # ---------------------------------------------------------------
    unitary_test()



#!/usr/bin/env python3
# Author:  Octavio Castillo Reyes
# Contact: octavio.castillo@bsc.es
"""Define classes and functions for parallel computations within **PETGEM**."""

# ---------------------------------------------------------------
# Load python modules
# ---------------------------------------------------------------
from mpi4py import MPI
from singleton_decorator import singleton
from petsc4py import PETSc

# ###############################################################
# ################     CLASSES DEFINITION      ##################
# ###############################################################

@singleton
class MPIEnvironment():
    """MPI class definition."""

    def __init__(self):
        """Class for initialization of an MPI environment.

        :param: None.
        :return: class for MPI environment.
        :rtype: mpi_env class.
        """
        # Store the MPI environment
        self.MPI = MPI

        # MPI communicator
        self.comm = MPI.COMM_WORLD
        # MPI process identifier
        self.rank = self.comm.Get_rank()
        # Size of the group associated with a communicator
        self.num_proc = self.comm.Get_size()
        # Machine name
        self.machine_name = self.MPI.Get_processor_name()

        return


# ###############################################################
# ################     FUNCTIONS DEFINITION     #################
# ###############################################################

def createSequentialDenseMatrixWithArray(dimension1, dimension2, data):
    """Given an input array, create a sequential dense matrix in petsc format.

    :param int dimension1: matrix dimension (rows).
    :param int dimension2: matrix dimension (columns).
    :param ndarray data: data to be exported.
    :return: parallel matrix.
    :rtype: petsc parallel and dense matrix.
    """
    parallel_matrix = PETSc.Mat().createDense([dimension1, dimension2],
                                              array=data, comm=PETSc.COMM_SELF)

    parallel_matrix.setFromOptions()
    parallel_matrix.setUp()

    return parallel_matrix


def writeParallelDenseMatrix(output_file, data, communicator=None):
    """Write a Petsc parallel dense matrix which format is defined by two files: output_file.dat and output_file.info.

    :param str output_file: file name to be saved.
    :param petsc matrix data: dense matrix to be saved.
    :param str communicator: mpi communicator.
    :return: None.
    """
    if communicator is None:
        communicator = PETSc.COMM_WORLD

    viewer = PETSc.Viewer().createBinary(output_file, mode='w', comm=communicator)
    viewer(data)

    return


def createSequentialVectorWithArray(data):
    """Given an input array, create a sequential vector in petsc format.

    :param ndarray data: data to be exported.
    :return: parallel matrix.
    :rtype: petsc parallel and dense matrix.
    """
    parallel_vector = PETSc.Vec().createWithArray(data, comm=PETSc.COMM_SELF)

    parallel_vector.setUp()

    return parallel_vector


def writePetscVector(output_file, data, communicator=None):
    """Write a Petsc vector which format is defined by two files: output_file.dat and output_file.info.

    :param str output_file: file name to be saved.
    :param petsc vector data: array to be saved.
    :param str communicator: mpi communicator.
    :return: None.
    """
    if communicator is None:
        communicator = PETSc.COMM_WORLD

    viewer = PETSc.Viewer().createBinary(output_file, mode='w',
                                         comm=communicator)
    viewer(data)

    return


def readPetscMatrix(input_file, communicator=None):
    """Read a Petsc matrix which format is defined by two files: input_file.dat and input_file.info.

    :param str input_file: file name to be readed.
    :param str communicator: mpi communicator.
    :return: petsc_matrix.
    :rtype: petsc sparse matrix.
    """
    if communicator is None:
        communicator = PETSc.COMM_WORLD

    viewer = PETSc.Viewer().createBinary(input_file, mode='r',
                                         comm=communicator)
    petsc_matrix = PETSc.Mat().load(viewer)

    return petsc_matrix


def readPetscVector(input_file, communicator=None):
    """Read a Petsc vector which format is defined by two files: input_file.dat and input_file.info.

    :param str input_file: file name to be readed.
    :param str communicator: mpi communicator.
    :return: petsc_vector.
    :rtype: petsc vector.
    """
    if communicator is None:
        communicator = PETSc.COMM_WORLD

    viewer = PETSc.Viewer().createBinary(input_file, mode='r',
                                         comm=communicator)
    petsc_vector = PETSc.Vec().load(viewer)

    return petsc_vector


def createParallelMatrix(dimension1,dimension2,nnz,matrix_type,communicator=None):
    """Create a parallel sparse matrix in petsc format.

    :param int dimension1: matrix dimension (rows).
    :param int dimension2: matrix dimension (columns).
    :param int nnz: not zero pattern for allocation.
    :param int matrix_type: matrix type for parallel computations.
    :param str communicator: mpi communicator.
    :return: parallel matrix.
    :rtype: petsc AIJ parallel matrix.
    """
    if communicator is None:
        communicator = PETSc.COMM_WORLD

    if matrix_type:    # Cuda support
        parallel_matrix = PETSc.Mat().create(comm=communicator)
        parallel_matrix.setType('aijcusparse')
        parallel_matrix.setSizes([dimension1, dimension2])
    elif not matrix_type:    # No cuda support
        parallel_matrix = PETSc.Mat().createAIJ([dimension1, dimension2], comm=communicator)
    else:
        raise ValueError('Cuda option=', matrix_type, ' not supported.')

    parallel_matrix.setPreallocationNNZ((nnz, nnz))
    parallel_matrix.setFromOptions()
    parallel_matrix.setUp()

    return parallel_matrix


def createParallelVector(size, vector_type, communicator=None):
    """Create a parallel vector in petsc format.

    :param int size: vector size.
    :param int vector_type: vector type for parallel computations.
    :param str communicator: mpi communicator.
    :return: parallel vector.
    :rtype: petsc parallel vector.
    """
    if communicator is None:
        communicator = PETSc.COMM_WORLD

    if vector_type:    # Cuda support
        parallel_vector = PETSc.Vec().create(comm=communicator)
        parallel_vector.setType('cuda')
        parallel_vector.setSizes(size)
    elif not vector_type:    # No cuda support
        parallel_vector = PETSc.Vec().createMPI(size, comm=communicator)
    else:
        raise ValueError('Cuda option=', vector_type, ' not supported.')

    parallel_vector.setUp()

    return parallel_vector


def createSequentialVector(size, vector_type, communicator=None):
    """Create a sequential vector in petsc format.

    :param int size: vector size.
    :param int vector_type: vector type for parallel computations.
    :param str communicator: mpi communicator.
    :return: sequential vector.
    :rtype: petsc sequential vector.
    """
    if communicator is None:
        communicator = PETSc.COMM_SELF

    if vector_type == 0:    # No cuda support
        sequential_vector = PETSc.Vec().createSeq(size, comm=communicator)
    elif vector_type == 1:    # Cuda support
        sequential_vector = PETSc.Vec().create(comm=communicator)
        sequential_vector.setType('cuda')
        sequential_vector.setSizes(size)

    sequential_vector.setUp()

    return sequential_vector


def createParallelDenseMatrix(dimension1, dimension2, communicator=None):
    """Create a parallel dense matrix in petsc format.

    :param int dimension1: matrix dimension (rows).
    :param int dimension2: matrix dimension (columns).
    :param str communicator: mpi communicator.
    :return: parallel matrix.
    :rtype: petsc parallel and dense matrix.
    """
    if communicator is None:
        communicator = PETSc.COMM_WORLD

    parallel_matrix = PETSc.Mat().createDense([dimension1, dimension2],
                                              comm=communicator)
    parallel_matrix.setFromOptions()
    parallel_matrix.setUp()

    return parallel_matrix


def writeDenseMatrix(output_file, data, communicator=None):
    """Write a Petsc dense matrix which format is defined by two files: output_file.dat and output_file.info.

    :param str output_file: file name to be saved.
    :param petsc matrix data: dense matrix to be saved.
    :param str communicator: mpi communicator.
    :return: None.
    """
    if communicator is None:
        communicator = PETSc.COMM_WORLD

    viewer = PETSc.Viewer().createBinary(output_file, mode='w',
                                         comm=communicator)
    viewer.pushFormat(viewer.Format.NATIVE)

    viewer(data)

    return


def unitary_test():
    """Unitary test for parallel.py script."""
# ###############################################################
# ################             MAIN             #################
# ###############################################################


if __name__ == '__main__':
    # ---------------------------------------------------------------
    # Run unitary test
    # ---------------------------------------------------------------
    unitary_test()



#!/usr/bin/env python3
# Author:  Octavio Castillo Reyes
# Contact: octavio.castillo@bsc.es
"""setup.py is the main petgem setup script, it is based on python setup-module."""


def name():
    """Set name for petgem package.

    :param: None
    :return: petgem name
    :rtype: string
    """
    return 'petgem'


def description():
    """Descripton of petgem code."""
    petgem_description = ('Parallel python code for electromagnetic '
                          'modeling in geophysics using high-order nédélec finite elements')

    return petgem_description


def long_description():
    """Read long description of petgem of DESCRIPTION.rst file.

    :param: None
    :return: petgem description
    :rtype: string
    """
    with open(os.path.join('DESCRIPTION.rst')) as f:
        return f.read()


def get_ext_modules():
    """Get paths of extension modules.

    :param: None
    :return: numpy path include
    """
    try:
        import numpy
        numpy_includes = [numpy.get_include()]
    except ImportError:
        numpy_includes = []

    return numpy_includes


if __name__ == '__main__':
    from setuptools import setup
    import os
    # petgem setup
    setup(name=name(),
          maintainer="Octavio Castillo-Reyes",
          maintainer_email="octavio.castillo@bsc.es",
          version='0.8',
          long_description=long_description(),
          description=description(),
          url="https://www.bsc.es/castillo-octavio",
          download_url="http://petgem.bsc.es",
          license='BSD-3',
          packages=['petgem'],
          include_dirs=get_ext_modules(),
          install_requires=['petsc4py', 'numpy', 'scipy', 'singleton_decorator', 'mpi4py', 'h5py'],
          setup_requires=['sphinx'],
          classifiers=['Development Status :: 3 - Alpha',
                       'License :: Free for non-commercial use',
                       'Intended Audience :: Science/Research',
                       'Intended Audience :: Developers',
                       'Programming Language :: Python :: 3.6',
                       'Topic :: Scientific/Engineering',
                       'Topic :: Software Development :: Libraries',
                       'Operating System :: POSIX :: Linux'],
          keywords=['3D CSEM, high-order edge finite elements, HPC, numerical simulation'],
          platforms="Linux",
          include_package_data=True,
          )

else:
    pass
